import sys

import click

from .conf import config
from .exceptions import ScreenError, StorageError
from .screen import take_image
from .storages import get_storage
from .utils import copy_to_clipboard, open_path
from .version import __version__

echo_error = lambda text: click.echo(click.style(text, fg='red'), err=True)


def prompt_config_value(ctx, param, value):
    if value is not None:
        return value

    if ctx.params['unset']:
        return None

    return click.prompt('Value', hide_input=True)


@click.group("cloudapp-screenshots")
@click.help_option('-h', '--help')
@click.version_option(__version__, '-v', '--version', message="v%(version)s")
def main():
    pass


@main.group('config', help="Get and set options.")
@click.help_option('-h', '--help')
def config_group():
    pass


@config_group.command('set', help="Set the option.")
@click.argument('key')
@click.argument('value', required=False, callback=prompt_config_value)
@click.help_option('-h', '--help')
@click.option('-U', '--unset', is_flag=True, default=False, is_eager=True, help="Remove the option.")
def config_set(key, value, unset):
    key = key.replace('.', '_').upper()

    if unset:
        delattr(config, key)
    else:
        setattr(config, key, value)


@config_group.command('reset', help="Remove all options.")
@click.help_option('-h', '--help')
@click.option('-f', '--force', is_flag=True, default=False)
def config_reset(force):
    if not force and not click.confirm("Are you sure?"):
        echo_error('Aborted!')
        sys.exit(1)

    config.reset(reload=False)


@config_group.command('list', help="Show options.")
@click.help_option('-h', '--help')
def config_list():
    lines = []

    prev_option = None
    for option in config:
        if not prev_option or prev_option.section != option.section:
            lines.append('[{}]'.format(option.section))
        lines.append("\t{} = '{}'".format(option.option, option.value))

        prev_option = option

    output = '\n'.join(lines)
    click.echo_via_pager(output)


@main.command('image', help="Make a screenshot and upload to a storage.")
@click.help_option('-h', '--help')
@click.option('-b', '--browse', 'is_browse', is_flag=True, help="Open a screenshot.")
@click.option('-c', '--clipboard', 'is_copy_to_clipboard', is_flag=True, help="Copy a screenshot path to clipboard.")
@click.option('-s', '--storage', 'storage_name', help="Choose a storage.")
def make_image(is_browse, is_copy_to_clipboard, storage_name):
    try:
        image_stream, fmt = take_image()
    except ScreenError as e:
        echo_error(e.message)
        sys.exit(1)

    try:
        storage = get_storage(storage_name)
        file_detail = storage.upload_image(image_stream, fmt)
    except StorageError as e:
        echo_error(e.message)
        sys.exit(1)

    if is_browse:
        open_path(file_detail.path)

    if is_copy_to_clipboard:
        copy_to_clipboard(file_detail.path)



class Error(Exception):
    default_message = "Oops, something went wrong"

    def __init__(self, message=None):
        super(Error, self).__init__(message)

        self.message = message or self.default_message

    def __str__(self):
        return self.message


class ScreenError(Error):
    pass


class StorageError(Error):
    pass



import logging
from io import BytesIO

from .grabber import grab_area

logger = logging.getLogger(__name__)

__all__ = ['take_image']


def take_image():
    logger.info("Taking an image.")

    coords = grab_area()
    stream, fmt = _capture_screen_via_mss(coords)

    return stream, fmt


def _capture_screen_via_mss(coords):
    from mss import mss
    from ._mss import to_png

    logger.info("Capturing a screen with mss.")

    monitor = {
        'top': coords[1],
        'left': coords[0],
        'width': coords[2] - coords[0] + 1,
        'height': coords[3] - coords[1] + 1,
    }

    screen = mss()
    image = screen.grab(monitor)

    stream = to_png(image.rgb, image.size)
    fmt = 'png'

    return stream, fmt


def _capture_screen_via_pyscreenshot(coords):
    import pyscreenshot

    logger.info("Capturing a screen with pyscreenshot.")

    stream = BytesIO()
    fmt = 'png'

    image = pyscreenshot.grab(coords)
    image.save(stream, format=fmt)

    return stream, fmt



import logging
import os
import time

import requests
from requests.auth import HTTPDigestAuth

from .base import BaseStorage, File
from ..conf import config
from ..exceptions import StorageError

logger = logging.getLogger(__name__)


class Storage(BaseStorage):
    base_url = 'https://my.cl.ly/v3'

    def __init__(self, username=config.CLOUDAPP_USERNAME, password=config.CLOUDAPP_PASSWORD):
        self._auth = HTTPDigestAuth(username, password)
        self._session = requests.Session()
        self._session.headers = {
            'Accept': 'application/json',
        }

    def upload_image(self, stream, fmt='png'):
        logger.info('Uploading an image to CloudApp.')
        filename = 'screenshot-%s.%s' % (time.time(), fmt)

        file_detail = self._upload_file(stream, filename)

        return File(File.IMAGE, file_detail['share_url'])

    def _upload_file(self, stream, filename):
        s3_meta = self._request('POST', 'items', data={'name': filename})

        s3_url = s3_meta['url']
        s3_data = s3_meta['s3']
        s3_files = {'file': (filename, stream)}

        logger.debug('Sending a POST request to %s: data=%s files=%s', s3_url, s3_data, s3_files)
        response = self._session.post(s3_url, data=s3_data, files=s3_files)

        return response.json()

    def _request(self, method, path, **kwargs):
        url = self._build_url(path)

        logger.debug('Sending a %s request to %s: %s', method, url, kwargs)
        try:
            response = self._session.request(method, url, auth=self._auth, **kwargs)
        except requests.RequestException as e:
            logger.error(e)
            raise StorageError("Can't send a request to CloudApp. {}".format(e))

        logger.debug('Request to %s failed: status=%s body=%s', url, response.status_code, response.text)

        if not response.ok:
            if response.status_code >= 500:
                error_msg = "CloudApp broken :("
            elif response.status_code == 400:
                error_msg = "CloudApp doesn't want to accept our request. Sorry :("
            elif response.status_code in (401, 403):
                error_msg = "Invalid credentials for CloudApp."
            else:
                error_msg = None
            raise StorageError(error_msg)

        return response.json()

    def _build_url(self, path):
        if path.startswith('/'):
            path = path[1:]

        return os.path.join(self.base_url, path)



"""
This is part of the MSS Python's module.
Source: https://github.com/BoboTiG/python-mss
License: MIT
2017-07-18
"""
import io
import struct
import zlib


def to_png(data, size):
    # type: (bytes, Tuple[int, int], str) -> None
    """
    Dump data to a PNG file.

    :param bytes data: RGBRGB...RGB data.
    :param tuple size: The (width, height) pair.
    """

    width, height = size
    line = width * 3
    png_filter = struct.pack('>B', 0)
    scanlines = b''.join(
        [png_filter + data[y * line:y * line + line]
         for y in range(height)])

    magic = struct.pack('>8B', 137, 80, 78, 71, 13, 10, 26, 10)

    # Header: size, marker, data, CRC32
    ihdr = [b'', b'IHDR', b'', b'']
    ihdr[2] = struct.pack('>2I5B', width, height, 8, 2, 0, 0, 0)
    ihdr[3] = struct.pack('>I', zlib.crc32(b''.join(ihdr[1:3])) & 0xffffffff)
    ihdr[0] = struct.pack('>I', len(ihdr[2]))

    # Data: size, marker, data, CRC32
    idat = [b'', b'IDAT', zlib.compress(scanlines), b'']
    idat[3] = struct.pack('>I', zlib.crc32(b''.join(idat[1:3])) & 0xffffffff)
    idat[0] = struct.pack('>I', len(idat[2]))

    # Footer: size, marker, None, CRC32
    iend = [b'', b'IEND', b'', b'']
    iend[3] = struct.pack('>I', zlib.crc32(iend[1]) & 0xffffffff)
    iend[0] = struct.pack('>I', len(iend[2]))

    # store data in memory
    fileh = io.BytesIO()
    fileh.write(magic)
    fileh.write(b''.join(ihdr))
    fileh.write(b''.join(idat))
    fileh.write(b''.join(iend))
    fileh.seek(0)

    return fileh



import logging.config
import os
from collections import namedtuple
from shutil import copyfile

from ..compat import PY2

if PY2:
    from ConfigParser import ConfigParser
else:
    from configparser import ConfigParser

logger = logging.getLogger(__name__)

LOGGER_NAMES = ('CRITICAL', 'ERROR', 'WARN', 'WARNING', 'INFO', 'DEBUG', 'NOTSET')

Option = namedtuple('Option', ('key', 'section', 'option', 'value'))


class Config(object):
    _path = os.path.join(os.path.expanduser('~'), '.config', 'grab-screen', 'config.ini')

    _auto_save = True

    _default_section = 'app'
    _sections = ('app', 'logger', 'storages', 'cloudapp')

    _default_option = 'default'

    def __init__(self):
        self._config = ConfigParser()

    def __iter__(self):
        for section in self._sections:
            if not self._config.has_section(section):
                continue

            for option, value in self._config.items(section):
                key = '{}_{}'.format(section, option).upper()
                yield Option(key, section, option, value)

    def __getattribute__(self, key):
        if key.isupper():
            return self._get_value(key)

        return object.__getattribute__(self, key)

    def __setattr__(self, key, value):
        if key.isupper():
            return self._set_value(key, value)

        return object.__setattr__(self, key, value)

    def __delattr__(self, key):
        if key.isupper():
            return self._del_value(key)

        return object.__delattr__(self, key)

    def load(self):
        if not os.path.exists(self._path):
            self.reset()

        self._config.read(self._path)

        self._load_logger()

    def save(self):
        self._create_path()

        with open(self._path, 'w') as config_file:
            self._config.write(config_file)

    def reset(self, reload=True):
        self._create_path()

        default_config_path = os.path.abspath(os.path.join(os.path.dirname(__file__), 'default.ini'))
        copyfile(default_config_path, self._path)

        if reload:
            self.load()

    def _create_path(self):
        directory = os.path.dirname(self._path)
        if not os.path.exists(directory):
            os.makedirs(directory)

    def _load_logger(self):
        level = self.LOGGER_LEVEL.upper()
        if level not in LOGGER_NAMES:
            logger.warning("Invalid logger level '%s'", level)
            level = 'INFO'

        logging.config.dictConfig({
            'version': 1,
            'formatters': {
                'simple': {
                    'format': '%(asctime)s [%(levelname)s] %(message)s'
                },
            },
            'handlers': {
                'console': {
                    'level': 'DEBUG',
                    'class': 'logging.StreamHandler',
                    'formatter': 'simple',
                },
            },
            'loggers': {
                'grab_screen': {
                    'level': self.LOGGER_LEVEL.upper(),
                    'handlers': ['console'],
                },
            }
        })

    def _get_value(self, key):
        key = key.lower()

        section, option = self._split_key(key)
        type_ = self._get_value_type(option)

        if not self._config.has_option(section, option):
            return None

        if type_ == 'int':
            getter = self._config.getint
        elif type_ == 'float':
            getter = self._config.getfloat
        elif type_ == 'boolean':
            getter = self._config.getboolean
        else:
            getter = self._config.get

        return getter(section, option)

    def _set_value(self, key, value):
        if value is None:
            return self._del_value(key)

        key = key.lower()

        section, option = self._split_key(key)

        self._config.set(section, option, str(value))

        if self._auto_save:
            self.save()

    def _del_value(self, key):
        key = key.lower()

        section, option = self._split_key(key)
        self._config.remove_option(section, option)

        if self._auto_save:
            self.save()

    def _split_key(self, key):
        words = key.split('_', 1)

        try:
            section = tuple(filter(lambda s: s == words[0], self._sections))[0]
            option = words[1] or self._default_option
        except IndexError:
            section = self._default_section
            option = key

        if not self._config.has_section(section):
            self._config.add_section(section)

        return section, option

    def _get_value_type(self, key):
        if key.startswith('is_'):
            return 'boolean'

        return ''



from setuptools import setup, find_packages

from grab_screen import __version__

# get the description from the README file
with open('README.md') as readme_file:
    long_description = readme_file.read()

# get the python dependencies
with open('requirements/main.txt') as req_file:
    requirements = req_file.readlines()

setup(
    name='grab-screen',
    description='Take screenshots and upload anywhere',
    long_description=long_description,
    version=__version__,
    license='MIT',
    url='https://github.com/andrei-shabanski/grab-screen',
    author='Andrei Shabanski',
    author_email='andrei.shabanski@gmail.com',
    packages=find_packages(exclude=['test']),
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'grab-screen=grab_screen:main'
        ]
    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Multimedia :: Graphics :: Capture',
        'Topic :: Multimedia :: Graphics :: Capture :: Screen Capture',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.0',
        'Programming Language :: Python :: 3.1',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='screenshot image capture grab cloudapp'
)



class File(object):
    IMAGE = 'image'
    GIF = 'gif'
    VIDEO = 'video'

    def __init__(self, file_type, path):
        self.file_type = file_type
        self.path = path


class BaseStorage(object):

    def upload_image(self, stream, fmt='png'):
        raise NotImplementedError()



import logging

from ..compat import PY2
from ..exceptions import ScreenError
from ..version import __name__

if PY2:
    import Tkinter as tk
else:
    import tkinter as tk

logger = logging.getLogger(__name__)


class Grabber(tk.Tk):
    WINDOW_COLOR = '#ffffff'
    WINDOW_ALPHA = 0.2

    RECTANGLE_COLOR = '#000000'

    @classmethod
    def run(cls):
        scope = {'coords': None}

        def set_coords(_coords):
            scope['coords'] = _coords
            grabber.exit()  # close the window

        grabber = cls(on_selected=set_coords)
        try:
            grabber.mainloop()
        except KeyboardInterrupt:
            grabber.exit()

        return scope['coords']

    def __init__(self, on_selected):
        tk.Tk.__init__(self)

        self.title(__name__)

        self._on_selected = on_selected

        self._coords = None
        self._rect_id = None
        self._is_drawing = False

        self._canvas = None

        self.initialize_geometry()
        self.initialize_controls()

    def initialize_geometry(self):
        self.wait_visibility()
        self.attributes('-topmost', True)
        self.attributes('-fullscreen', True)
        self.attributes('-alpha', self.WINDOW_ALPHA)

    def initialize_controls(self):
        self._canvas = tk.Canvas(self, bg=self.WINDOW_COLOR, cursor='crosshair')
        self._canvas.pack(fill=tk.BOTH, expand=1)

        self._canvas.bind('<Button-1>', self.start_drawing)
        self._canvas.bind('<Button-3>', self.exit)
        self._canvas.bind('<ButtonRelease-1>', self.stop_drawing)
        self._canvas.bind('<Motion>', self.draw_rectangle)

    def start_drawing(self, event):
        logger.debug("Selecting screen area.")
        self._is_drawing = True

        x = self._canvas.canvasx(event.x)
        y = self._canvas.canvasy(event.y)

        self._coords = [x, y, x, y]
        self._rect_id = self._canvas.create_rectangle(*self._coords, fill=self.RECTANGLE_COLOR)

    def draw_rectangle(self, event):
        if not self._is_drawing:
            return

        self._coords[2] = self._canvas.canvasx(event.x)
        self._coords[3] = self._canvas.canvasy(event.y)
        self._canvas.coords(self._rect_id, *self._coords)

    def stop_drawing(self, event):
        logger.debug("Screen area has been selected. Coords: %s", self._coords)
        self._is_drawing = False

        self._canvas.delete(self._rect_id)

        self._on_selected(self._coords)

    def exit(self, event=None):
        self.attributes('-alpha', 0)
        self.destroy()


def grab_area():
    logger.info("Selecting an area.")
    coords = Grabber.run()
    if not coords:
        raise ScreenError("Aborted!")

    coords = tuple(map(int, coords))
    x1, x2 = sorted(coords[0::2])
    y1, y2 = sorted(coords[1::2])
    coords = (x1, y1, x2, y2)

    logger.debug("Selected area %s.", coords)
    return coords
